@@ -34,6 +34,7 @@ class Russound:
         self._callbacks = {}
         self._connection_started = False
         self._watched_devices = {}
+        self._controllers = {}
         self.rio_version = None
 
     def _retrieve_cached_variable(self, device_str: str, key: str):
@@ -63,6 +64,14 @@ class Russound:
         # Handle callbacks
         for callback in self._callbacks.get(device_str, []):
             callback(device_str, key, value)
+        # Handle source callback
+        if device_str[0] is 'S':
+            for controller in self._controllers.values():
+                for zone in controller.zones.values():
+                    source = zone.fetch_current_source()
+                    if source and source.device_str() is device_str:
+                        for callback in self._callbacks.get(zone.device_str(), []):
+                            callback(device_str, key, value)
 
     def _process_response(self, res: bytes):
         s = str(res, "utf-8").strip()
@@ -86,9 +95,17 @@ class Russound:
 
         return ty, p["value"] or p["value_only"]
 
+    async def _keep_alive(self):
+        while True:
+            await asyncio.sleep(900)  # 15 minutes
+            _LOGGER.debug("Sending keep alive to device")
+            await self._send_cmd("VERSION")
+
     async def _ioloop(self, reader: StreamReader, writer: StreamWriter, reconnect: bool):
         queue_future = ensure_future(self._cmd_queue.get())
         net_future = ensure_future(reader.readline())
+        keep_alive_task = asyncio.create_task(self._keep_alive())
+
         try:
             _LOGGER.debug("Starting IO loop")
             while True:
@@ -125,9 +142,6 @@ class Russound:
                             break
         except asyncio.CancelledError:
             _LOGGER.debug("IO loop cancelled")
-            writer.close()
-            queue_future.cancel()
-            net_future.cancel()
             raise
         except asyncio.TimeoutError:
             _LOGGER.warning("Connection to Russound client timed out")
@@ -137,12 +151,18 @@ class Russound:
             _LOGGER.exception("Unhandled exception in IO loop")
             raise
         finally:
+            _LOGGER.debug("Cancelling all tasks...")
+            writer.close()
+            queue_future.cancel()
+            net_future.cancel()
+            keep_alive_task.cancel()
             if reconnect and self._connection_started:
                 _LOGGER.info("Retrying connection to Russound client in 5s")
                 await asyncio.sleep(RECONNECT_DELAY)
                 await self.connect(reconnect)
 
     async def _send_cmd(self, cmd: str):
+        _LOGGER.debug(f"Sending command '{cmd}' to Russound client")
         future = asyncio.Future()
         await self._cmd_queue.put((cmd, future))
         r = await future
@@ -217,9 +237,9 @@ class Russound:
 
     async def enumerate_controllers(self):
         """Return a list of (controller_id, controller_macAddress, controller_type) tuples"""
-        controllers: list[Controller] = []
+        controllers: dict[int, Controller] = {}
         # Search for first controller, then iterate if RNET is supported
-        for controller_id in range(1, 8):
+        for controller_id in range(1, 9):
             device_str = controller_device_str(controller_id)
             try:
                 mac_address = await self.get_variable(
@@ -237,12 +257,13 @@ class Russound:
                     firmware_version = await self.get_variable(
                         device_str, "firmwareVersion"
                     )
-                controller = Controller(self, controller_id, mac_address, controller_type, firmware_version)
+                controller = Controller(self, controllers.get(1, None), controller_id, mac_address, controller_type,
+                                        firmware_version)
                 await controller.fetch_configuration()
-                controllers.append(controller)
+                controllers[controller_id] = controller
             except CommandException:
                 continue
-
+        self._controllers = controllers
         return controllers
 
     @property
@@ -271,9 +292,10 @@ class Russound:
 class Controller:
     """Uniquely identifies a controller"""
 
-    def __init__(self, instance: Russound, controller_id: int, mac_address: str, controller_type: str,
-                 firmware_version: str):
+    def __init__(self, instance: Russound, parent_controller, controller_id: int, mac_address: str,
+                 controller_type: str, firmware_version: str):
         self.instance = instance
+        self.parent_controller = parent_controller
         self.controller_id = controller_id
         self.mac_address = mac_address
         self.controller_type = controller_type
@@ -302,7 +324,7 @@ class Controller:
     async def _init_zones(self):
         """Return a list of (zone_id, zone) tuples"""
         self.zones = {}
-        for zone_id in range(1, self.max_zones):
+        for zone_id in range(1, self.max_zones + 1):
             try:
                 device_str = zone_device_str(self.controller_id, zone_id)
                 name = await self.instance.get_variable(device_str, "name")
@@ -417,7 +439,7 @@ class Zone:
 
     @property
     def volume(self):
-        return self._get('volume')
+        return self._get('volume', '0')
 
     @property
     def bass(self):
@@ -449,10 +471,10 @@ class Zone:
 
     @property
     def status(self):
-        return self._get('status')
+        return self._get('status', 'OFF')
 
     @property
-    def mute(self):
+    def is_mute(self):
         return self._get('mute')
 
     @property
@@ -479,6 +501,30 @@ class Zone:
     def enabled(self):
         return self._get('enabled')
 
+    async def mute(self):
+        return await self.send_event('ZoneMuteOn')
+
+    async def unmute(self):
+        return await self.send_event('ZoneMuteOff')
+
+    async def set_volume(self, volume):
+        return await self.send_event('KeyPress', 'Volume', volume)
+
+    async def volume_up(self):
+        return await self.send_event('KeyPress', 'VolumeUp')
+
+    async def volume_down(self):
+        return await self.send_event('KeyPress', 'VolumeDown')
+
+    async def zone_on(self):
+        return await self.send_event('ZoneOn')
+
+    async def zone_off(self):
+        return await self.send_event('ZoneOff')
+
+    async def select_source(self, source: int):
+        return await self.send_event('SelectSource', source)
+
 
 class Source:
     """Uniquely identifies a Source"""
